@@ -17,19 +17,19 @@ def test_install_layoutparser(root_node):
         ), "Install layoutparser Failed"
 
 
-def test_parse_layout(pdf_nodes):
+def test_parse_layout(app, pdf_nodes):
     """
     Check that the layout model runs and adds content properly
     """
     doc_node = pdf_nodes[0]
-    doc_node.nodes = []  # clear any previously defined nodes
+    app.tree.remove_children(doc_node)
 
     AutoTools(doc_node).extract_layout()
-    assert len(doc_node.nodes) > 0, "Layoutparser could not find any nodes"
+    assert len(doc_node.data['children']) > 0, "Layoutparser could not find any nodes"
 
     types = defaultdict(list)
-    for x in doc_node.dfs():
-        types[x._type].append(x.label)
+    for x in app.tree.dfs(doc_node.id):
+        types[x.data['type']].append(x.data.get('label',''))
 
     assert len(types["section"]) == 3, types["section"]
     # Finding the image is too flaky
